# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("p2pgraph.dll")
prototypes = \
    {
        #
        'PeerGraphStartup': SimTypeFunction([SimTypeShort(signed=False, label="UInt16"), SimTypePointer(SimStruct({"wVersion": SimTypeShort(signed=False, label="UInt16"), "wHighestVersion": SimTypeShort(signed=False, label="UInt16")}, name="PEER_VERSION_DATA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["wVersionRequested", "pVersionData"]),
        #
        'PeerGraphShutdown': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        #
        'PeerGraphFreeData': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeBottom(label="Void"), arg_names=["pvData"]),
        #
        'PeerGraphGetItemCount': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hPeerEnum", "pCount"]),
        #
        'PeerGraphGetNextItem': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hPeerEnum", "pCount", "pppvItems"]),
        #
        'PeerGraphEndEnumeration': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hPeerEnum"]),
        #
        'PeerGraphCreate': SimTypeFunction([SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwScope": SimTypeInt(signed=False, label="UInt32"), "dwMaxRecordSize": SimTypeInt(signed=False, label="UInt32"), "pwzGraphId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzFriendlyName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ulPresenceLifetime": SimTypeInt(signed=False, label="UInt32"), "cPresenceMax": SimTypeInt(signed=False, label="UInt32")}, name="PEER_GRAPH_PROPERTIES", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "pwzSspFilename": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzPackageName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cbSecurityInfo": SimTypeInt(signed=False, label="UInt32"), "pbSecurityInfo": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pvContext": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "pfnValidateRecord": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "type": SimTypeBottom(label="Guid"), "id": SimTypeBottom(label="Guid"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzModifiedById": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzAttributes": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ftCreation": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftExpiration": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftLastModified": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "securityData": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), "data": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None)}, name="PEER_RECORD", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="PEER_RECORD_CHANGE_TYPE")], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pvContext", "pRecord", "changeType"]), offset=0), "pfnSecureRecord": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "type": SimTypeBottom(label="Guid"), "id": SimTypeBottom(label="Guid"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzModifiedById": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzAttributes": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ftCreation": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftExpiration": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftLastModified": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "securityData": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), "data": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None)}, name="PEER_RECORD", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="PEER_RECORD_CHANGE_TYPE"), SimTypePointer(SimTypePointer(SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pvContext", "pRecord", "changeType", "ppSecurityData"]), offset=0), "pfnFreeSecurityData": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pvContext", "pSecurityData"]), offset=0), "pfnAuthFailed": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pvContext"]), offset=0)}, name="PEER_SECURITY_INTERFACE", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pGraphProperties", "pwzDatabaseName", "pSecurityInterface", "phGraph"]),
        #
        'PeerGraphOpen': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "pwzSspFilename": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzPackageName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cbSecurityInfo": SimTypeInt(signed=False, label="UInt32"), "pbSecurityInfo": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "pvContext": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "pfnValidateRecord": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "type": SimTypeBottom(label="Guid"), "id": SimTypeBottom(label="Guid"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzModifiedById": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzAttributes": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ftCreation": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftExpiration": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftLastModified": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "securityData": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), "data": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None)}, name="PEER_RECORD", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="PEER_RECORD_CHANGE_TYPE")], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pvContext", "pRecord", "changeType"]), offset=0), "pfnSecureRecord": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "type": SimTypeBottom(label="Guid"), "id": SimTypeBottom(label="Guid"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzModifiedById": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzAttributes": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ftCreation": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftExpiration": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftLastModified": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "securityData": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), "data": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None)}, name="PEER_RECORD", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="PEER_RECORD_CHANGE_TYPE"), SimTypePointer(SimTypePointer(SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pvContext", "pRecord", "changeType", "ppSecurityData"]), offset=0), "pfnFreeSecurityData": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pvContext", "pSecurityData"]), offset=0), "pfnAuthFailed": SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pvContext"]), offset=0)}, name="PEER_SECURITY_INTERFACE", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwzGraphId", "pwzPeerId", "pwzDatabaseName", "pSecurityInterface", "cRecordTypeSyncPrecedence", "pRecordTypeSyncPrecedence", "phGraph"]),
        #
        'PeerGraphListen': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeShort(signed=False, label="UInt16")], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "dwScope", "dwScopeId", "wPort"]),
        #
        'PeerGraphConnect': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "sin6": SimStruct({"sin6_family": SimTypeShort(signed=False, label="UInt16"), "sin6_port": SimTypeShort(signed=False, label="UInt16"), "sin6_flowinfo": SimTypeInt(signed=False, label="UInt32"), "sin6_addr": SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), "Anonymous": SimUnion({"sin6_scope_id": SimTypeInt(signed=False, label="UInt32"), "sin6_scope_struct": SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Value": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="SCOPE_ID", pack=False, align=None)}, name="<anon>", label="None")}, name="SOCKADDR_IN6", pack=False, align=None)}, name="PEER_ADDRESS", pack=False, align=None), offset=0), SimTypePointer(SimTypeLongLong(signed=False, label="UInt64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pwzPeerId", "pAddress", "pullConnectionId"]),
        #
        'PeerGraphClose': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph"]),
        #
        'PeerGraphDelete': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pwzGraphId", "pwzPeerId", "pwzDatabaseName"]),
        #
        'PeerGraphGetStatus': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pdwStatus"]),
        #
        'PeerGraphGetProperties': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwScope": SimTypeInt(signed=False, label="UInt32"), "dwMaxRecordSize": SimTypeInt(signed=False, label="UInt32"), "pwzGraphId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzFriendlyName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ulPresenceLifetime": SimTypeInt(signed=False, label="UInt32"), "cPresenceMax": SimTypeInt(signed=False, label="UInt32")}, name="PEER_GRAPH_PROPERTIES", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "ppGraphProperties"]),
        #
        'PeerGraphSetProperties': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwScope": SimTypeInt(signed=False, label="UInt32"), "dwMaxRecordSize": SimTypeInt(signed=False, label="UInt32"), "pwzGraphId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzFriendlyName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzComment": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ulPresenceLifetime": SimTypeInt(signed=False, label="UInt32"), "cPresenceMax": SimTypeInt(signed=False, label="UInt32")}, name="PEER_GRAPH_PROPERTIES", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pGraphProperties"]),
        #
        'PeerGraphRegisterEvent': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"eventType": SimTypeInt(signed=False, label="PEER_GRAPH_EVENT_TYPE"), "pType": SimTypePointer(SimTypeBottom(label="Guid"), offset=0)}, name="PEER_GRAPH_EVENT_REGISTRATION", pack=False, align=None), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "hEvent", "cEventRegistrations", "pEventRegistrations", "phPeerEvent"]),
        #
        'PeerGraphUnregisterEvent': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hPeerEvent"]),
        #
        'PeerGraphGetEventData': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"eventType": SimTypeInt(signed=False, label="PEER_GRAPH_EVENT_TYPE"), "Anonymous": SimUnion({"dwStatus": SimTypeInt(signed=False, label="PEER_GRAPH_STATUS_FLAGS"), "incomingData": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "ullConnectionId": SimTypeLongLong(signed=False, label="UInt64"), "type": SimTypeBottom(label="Guid"), "data": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None)}, name="PEER_EVENT_INCOMING_DATA", pack=False, align=None), "recordChangeData": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "changeType": SimTypeInt(signed=False, label="PEER_RECORD_CHANGE_TYPE"), "recordId": SimTypeBottom(label="Guid"), "recordType": SimTypeBottom(label="Guid")}, name="PEER_EVENT_RECORD_CHANGE_DATA", pack=False, align=None), "connectionChangeData": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "status": SimTypeInt(signed=False, label="PEER_CONNECTION_STATUS"), "ullConnectionId": SimTypeLongLong(signed=False, label="UInt64"), "ullNodeId": SimTypeLongLong(signed=False, label="UInt64"), "ullNextConnectionId": SimTypeLongLong(signed=False, label="UInt64"), "hrConnectionFailedReason": SimTypeInt(signed=True, label="Int32")}, name="PEER_EVENT_CONNECTION_CHANGE_DATA", pack=False, align=None), "nodeChangeData": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "changeType": SimTypeInt(signed=False, label="PEER_NODE_CHANGE_TYPE"), "ullNodeId": SimTypeLongLong(signed=False, label="UInt64"), "pwzPeerId": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="PEER_EVENT_NODE_CHANGE_DATA", pack=False, align=None), "synchronizedData": SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "recordType": SimTypeBottom(label="Guid")}, name="PEER_EVENT_SYNCHRONIZED_DATA", pack=False, align=None)}, name="<anon>", label="None")}, name="PEER_GRAPH_EVENT_DATA", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hPeerEvent", "ppEventData"]),
        #
        'PeerGraphGetRecord': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "type": SimTypeBottom(label="Guid"), "id": SimTypeBottom(label="Guid"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzModifiedById": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzAttributes": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ftCreation": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftExpiration": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftLastModified": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "securityData": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), "data": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None)}, name="PEER_RECORD", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pRecordId", "ppRecord"]),
        #
        'PeerGraphAddRecord': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "type": SimTypeBottom(label="Guid"), "id": SimTypeBottom(label="Guid"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzModifiedById": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzAttributes": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ftCreation": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftExpiration": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftLastModified": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "securityData": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), "data": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None)}, name="PEER_RECORD", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pRecord", "pRecordId"]),
        #
        'PeerGraphUpdateRecord': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "type": SimTypeBottom(label="Guid"), "id": SimTypeBottom(label="Guid"), "dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "pwzCreatorId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzModifiedById": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwzAttributes": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ftCreation": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftExpiration": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "ftLastModified": SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), "securityData": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None), "data": SimStruct({"cbData": SimTypeInt(signed=False, label="UInt32"), "pbData": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="PEER_DATA", pack=False, align=None)}, name="PEER_RECORD", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pRecord"]),
        #
        'PeerGraphDeleteRecord': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pRecordId", "fLocal"]),
        #
        'PeerGraphEnumRecords': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pRecordType", "pwzPeerId", "phPeerEnum"]),
        #
        'PeerGraphSearchRecords': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pwzCriteria", "phPeerEnum"]),
        #
        'PeerGraphExportDatabase': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pwzFilePath"]),
        #
        'PeerGraphImportDatabase': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pwzFilePath"]),
        #
        'PeerGraphValidateDeferredRecords': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "cRecordIds", "pRecordIds"]),
        #
        'PeerGraphOpenDirectConnection': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "sin6": SimStruct({"sin6_family": SimTypeShort(signed=False, label="UInt16"), "sin6_port": SimTypeShort(signed=False, label="UInt16"), "sin6_flowinfo": SimTypeInt(signed=False, label="UInt32"), "sin6_addr": SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), "Anonymous": SimUnion({"sin6_scope_id": SimTypeInt(signed=False, label="UInt32"), "sin6_scope_struct": SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Value": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="SCOPE_ID", pack=False, align=None)}, name="<anon>", label="None")}, name="SOCKADDR_IN6", pack=False, align=None)}, name="PEER_ADDRESS", pack=False, align=None), offset=0), SimTypePointer(SimTypeLongLong(signed=False, label="UInt64"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pwzPeerId", "pAddress", "pullConnectionId"]),
        #
        'PeerGraphSendData': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "ullConnectionId", "pType", "cbData", "pvData"]),
        #
        'PeerGraphCloseDirectConnection': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeLongLong(signed=False, label="UInt64")], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "ullConnectionId"]),
        #
        'PeerGraphEnumConnections': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "dwFlags", "phPeerEnum"]),
        #
        'PeerGraphEnumNodes': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pwzPeerId", "phPeerEnum"]),
        #
        'PeerGraphSetPresence': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "fPresent"]),
        #
        'PeerGraphGetNodeInfo': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "ullNodeId": SimTypeLongLong(signed=False, label="UInt64"), "pwzPeerId": SimTypePointer(SimTypeChar(label="Char"), offset=0), "cAddresses": SimTypeInt(signed=False, label="UInt32"), "pAddresses": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "sin6": SimStruct({"sin6_family": SimTypeShort(signed=False, label="UInt16"), "sin6_port": SimTypeShort(signed=False, label="UInt16"), "sin6_flowinfo": SimTypeInt(signed=False, label="UInt32"), "sin6_addr": SimStruct({"u": SimUnion({"Byte": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 16), "Word": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 8)}, name="<anon>", label="None")}, name="IN6_ADDR", pack=False, align=None), "Anonymous": SimUnion({"sin6_scope_id": SimTypeInt(signed=False, label="UInt32"), "sin6_scope_struct": SimStruct({"Anonymous": SimUnion({"Anonymous": SimStruct({"_bitfield": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Value": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None")}, name="SCOPE_ID", pack=False, align=None)}, name="<anon>", label="None")}, name="SOCKADDR_IN6", pack=False, align=None)}, name="PEER_ADDRESS", pack=False, align=None), offset=0), "pwzAttributes": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="PEER_NODE_INFO", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "ullNodeId", "ppNodeInfo"]),
        #
        'PeerGraphSetNodeAttributes': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pwzAttributes"]),
        #
        'PeerGraphPeerTimeToUniversalTime': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pftPeerTime", "pftUniversalTime"]),
        #
        'PeerGraphUniversalTimeToPeerTime': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"dwLowDateTime": SimTypeInt(signed=False, label="UInt32"), "dwHighDateTime": SimTypeInt(signed=False, label="UInt32")}, name="FILETIME", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hGraph", "pftUniversalTime", "pftPeerTime"]),
    }

lib.set_prototypes(prototypes)



# pylint:disable=missing-class-docstring
from typing import List, Tuple, Any, Optional, Union, OrderedDict as ODict

import claripy
import ailment
import ailment.utils


INDENT_DELTA = 2


class EmptyBlockNotice(Exception):
    pass


class MultiNode:
    __slots__ = (
        "nodes",
        "addr",
        "idx",
    )

    def __init__(self, nodes, addr=None, idx=None):
        # delayed import
        from ..graph_region import GraphRegion  # pylint:disable=import-outside-toplevel

        self.nodes = []

        for node in nodes:
            if type(node) is MultiNode:
                self.nodes += node.nodes
            elif type(node) is GraphRegion:
                self.nodes += node.nodes
            else:
                self.nodes.append(node)

        self.addr = addr if addr is not None else self.nodes[0].addr
        self.idx = idx if idx is not None else self.nodes[0].idx if isinstance(self.nodes[0], ailment.Block) else None

    def copy(self):
        return MultiNode(self.nodes[::], addr=self.addr, idx=self.idx)

    def __repr__(self):
        addrs = []
        s = ""
        for node in self.nodes:
            if hasattr(node, "addr"):
                addrs.append(node.addr)
            s = f": {min(addrs):#x}-{max(addrs):#x}"

        return "<MultiNode %#x of %d nodes%s>" % (self.addr, len(self.nodes), s)

    def __hash__(self):
        # changing self.nodes does not change the hash, which enables in-place editing
        return hash((MultiNode, self.addr, self.idx))

    def __eq__(self, other):
        return isinstance(other, MultiNode) and self.nodes == other.nodes

    def dbg_repr(self, indent=0):
        s = ""
        for node in self.nodes:
            s += node.dbg_repr(indent=indent + INDENT_DELTA)
            s += "\n"

        return s


class BaseNode:
    __slots__ = ()

    @staticmethod
    def test_empty_node(node):
        # pylint:disable=simplifiable-if-statement
        if type(node) is ailment.Block:
            if not node.statements:
                return True
            # not empty
            return False
        elif type(node) is CodeNode:
            return BaseNode.test_empty_node(node.node)
        # unsupported node type. probably not empty?
        return False

    @staticmethod
    def test_empty_condition_node(cond_node):
        for node in [cond_node.true_node, cond_node.false_node]:
            if node is None:
                continue
            if type(node) is CodeNode and BaseNode.test_empty_node(node.node):
                continue
            if BaseNode.test_empty_node(node):
                continue
            return False

        return True

    addr: Optional[int]

    def dbg_repr(self, indent=0):
        return " " * indent + f"## dbg_repr not implemented for {type(self).__name__}"


class SequenceNode(BaseNode):
    __slots__ = (
        "addr",
        "nodes",
    )

    def __init__(self, addr: Optional[int], nodes=None):
        self.addr = addr
        self.nodes = nodes if nodes is not None else []

    def __repr__(self):
        if self.addr is None:
            return "<SequenceNode, %d nodes>" % len(self.nodes)
        else:
            return "<SequenceNode %#x, %d nodes>" % (self.addr, len(self.nodes))

    def add_node(self, node):
        self.nodes.append(node)

    def insert_node(self, pos, node):
        self.nodes.insert(pos, node)

    def remove_node(self, node):
        self.nodes.remove(node)

    def node_position(self, node):
        return self.nodes.index(node)

    def copy(self):
        return SequenceNode(self.addr, nodes=self.nodes[::])

    def dbg_repr(self, indent=0):
        s = ""
        for node in self.nodes:
            s += node.dbg_repr(indent=indent + INDENT_DELTA)
            s += "\n"

        return s


class CodeNode(BaseNode):
    __slots__ = (
        "node",
        "reaching_condition",
    )

    def __init__(self, node, reaching_condition):
        self.node = node
        self.reaching_condition = reaching_condition

    def __repr__(self):
        if self.addr is not None:
            if self.idx is not None:
                return f"<CodeNode {self.addr:#x}.{self.idx}>"
            return "<CodeNode %#x>" % self.addr
        else:
            return "<CodeNode %s>" % repr(self.node)

    @property
    def addr(self):
        if hasattr(self.node, "addr"):
            return self.node.addr
        else:
            return None

    @property
    def idx(self):
        if hasattr(self.node, "idx"):
            return self.node.idx
        return None

    def dbg_repr(self, indent=0):
        indent_str = indent * " "
        s = ""
        if self.reaching_condition is not None and not claripy.is_true(self.reaching_condition):
            s += (
                indent_str
                + "if (<block-missing>; %s)\n"
                + indent_str
                + "{\n"
                + indent_str
                + "  %s\n"
                + indent_str
                + "}"
            ) % (self.reaching_condition, self.node)
        else:
            s += indent_str + str(self.node)

        return s

    def copy(self):
        return CodeNode(self.node, self.reaching_condition)


class ConditionNode(BaseNode):
    __slots__ = (
        "addr",
        "node",
        "reaching_condition",
        "condition",
        "true_node",
        "false_node",
    )

    def __init__(self, addr, reaching_condition, condition, true_node, false_node=None):
        self.addr = addr
        self.reaching_condition = reaching_condition
        self.condition = condition
        self.true_node = true_node
        self.false_node = false_node

    def dbg_repr(self, indent=0):
        indent_str = indent * " "
        s = (
            indent_str + f"if (<block-missing>; {self.condition})\n{indent_str}{{\n"
            f"{self.true_node.dbg_repr(indent + INDENT_DELTA) if self.true_node is not None else ''}{indent_str}}}\n"
        )
        if self.false_node is not None:
            s += f"{indent_str}else\n{indent_str}{{\n{self.false_node.dbg_repr(indent + INDENT_DELTA)}{indent_str}}}\n"

        return s

    def __repr__(self):
        if self.addr is not None:
            return "<ConditionNode %#x>" % self.addr
        else:
            return f"<ConditionNode ({self.true_node!r}|{self.false_node!r})>"


class CascadingConditionNode(BaseNode):
    __slots__ = (
        "addr",
        "condition_and_nodes",
        "else_node",
    )

    def __init__(self, addr, condition_and_nodes: List[Tuple[Any, BaseNode]], else_node: BaseNode = None):
        self.addr = addr
        self.condition_and_nodes = condition_and_nodes
        self.else_node = else_node


class LoopNode(BaseNode):
    __slots__ = (
        "sort",
        "condition",
        "sequence_node",
        "initializer",
        "iterator",
        "_addr",
        "_continue_addr",
    )

    def __init__(self, sort, condition, sequence_node, addr=None, continue_addr=None, initializer=None, iterator=None):
        self.sort = sort
        self.condition = condition
        self.sequence_node = sequence_node
        self.initializer = initializer
        self.iterator = iterator
        self._addr = addr
        self._continue_addr = continue_addr

    def copy(self):
        return LoopNode(
            self.sort,
            self.condition,
            self.sequence_node,
            addr=self._addr,
            initializer=self.initializer,
            iterator=self.iterator,
        )

    @property
    def addr(self):
        if self._addr is None:
            return self.sequence_node.addr
        else:
            return self._addr

    @property
    def continue_addr(self):
        if self._continue_addr is None:
            return self.addr
        else:
            return self._continue_addr

    @continue_addr.setter
    def continue_addr(self, value):
        self._continue_addr = value

    def __repr__(self):
        return f"<LoopNode {self.sort}@{self.addr:#x}>"

    def dbg_repr(self, indent=0):
        # initializer = self.initializer.dbg_repr() if self.initializer is not None else 'None'
        # iterator = self.iterator.dbg_repr() if self.iterator else 'None'
        addr = hex(self.addr) if isinstance(self.addr, int) else str(self.addr)
        continue_addr = hex(self.continue_addr) if isinstance(self.continue_addr, int) else str(self.continue_addr)
        indent_str = " " * indent
        return (
            f"{indent_str}LoopNode(sort={self.sort}, initializer={self.initializer}, condition={self.condition}, "
            f"iterator={self.iterator}, addr={addr}, continue_addr={continue_addr}):\n"
            f"{self.sequence_node.dbg_repr(indent=indent + INDENT_DELTA)}"
        )


class BreakNode(BaseNode):
    __slots__ = (
        "addr",
        "target",
    )

    def __init__(self, addr, target):
        self.addr = addr
        self.target = target

    def dbg_repr(self, indent=0):
        return " " * indent + "BreakNode"


class ContinueNode(BaseNode):
    __slots__ = (
        "addr",
        "target",
    )

    def __init__(self, addr, target):
        self.addr = addr
        self.target = target

    def dbg_repr(self, indent=0):
        return " " * indent + "ContinueNode"


class ConditionalBreakNode(BreakNode):
    __slots__ = ("condition",)

    def __init__(self, addr, condition, target):
        super().__init__(addr, target)
        self.condition = condition

    def __repr__(self):
        return f"<ConditionalBreakNode {self.addr:#x} target:{self.target}>"

    def dbg_repr(self, indent=0):
        return " " * indent + "ConditionalBreakNode(condition={self.condition})"


class SwitchCaseNode(BaseNode):
    __slots__ = (
        "switch_expr",
        "cases",
        "default_node",
        "addr",
    )

    def __init__(self, switch_expr, cases: ODict[Union[int, Tuple[int, ...]], SequenceNode], default_node, addr=None):
        self.switch_expr = switch_expr
        self.cases: ODict[Union[int, Tuple[int, ...]], SequenceNode] = cases
        self.default_node = default_node
        self.addr = addr


class IncompleteSwitchCaseNode(BaseNode):
    """
    Describes an incomplete set of switch-case nodes. Usually an intermediate result. Should always be restructured
    into a SwitchCaseNode by the end of structuring. Only used in Phoenix structurer.
    """

    __slots__ = ("addr", "head", "cases")

    def __init__(self, addr, head, cases: List):
        self.addr = addr
        self.head = head
        self.cases: List = cases


#
# The following classes are custom AIL statements (not nodes, unfortunately)
#


class IncompleteSwitchCaseHeadStatement(ailment.statement.Statement):
    """
    Describes a switch-case head. This is only created by LoweredSwitchSimplifier.
    """

    __slots__ = ("addr", "switch_variable", "case_addrs", "_case_addrs_str")

    def __init__(self, idx, switch_variable, case_addrs, **kwargs):
        super().__init__(idx, **kwargs)
        self.switch_variable = switch_variable
        # original cmp node, case value | "default", address of the case node, idx of the case node,
        # address of the next cmp node
        self.case_addrs: List[Tuple[ailment.Block, Union[int, str], int, Optional[int], int]] = case_addrs
        # a string representation of the addresses of all cases, used for hashing
        self._case_addrs_str = str(sorted([c[0].addr for c in self.case_addrs if c[0] is not None]))

    def __repr__(self):
        return f"SwitchCaseHead: switch {self.switch_variable} with {len(self.case_addrs)} cases"

    def __str__(self):
        return f"switch ({str(self.switch_variable)}): {len(self.case_addrs)} cases"

    __hash__ = ailment.statement.TaggedObject.__hash__

    def _hash_core(self):
        return ailment.utils.stable_hash(
            (IncompleteSwitchCaseHeadStatement, self.idx, self.switch_variable, self._case_addrs_str)
        )



import angr


# pylint:disable=redefined-builtin,arguments-differ
class syscall(angr.SimProcedure):
    def run(self, *args, resolves=None):
        self.resolves = resolves  # pylint:disable=attribute-defined-outside-init

        if self.successors is not None:
            self.successors.artifacts["resolves"] = resolves

        # code duplicated from ReturnUnconstrained
        size = self.prototype.returnty.size
        # ummmmm do we really want to rely on this behavior?
        if size is NotImplemented:
            return None
        else:
            return self.state.solver.Unconstrained(
                "syscall_stub_%s" % self.display_name, size, key=("syscall", "?", self.display_name)
            )

    def __repr__(self):
        if "resolves" in self.kwargs:
            return "<Syscall stub (%s)>" % self.kwargs["resolves"]
        else:
            return "<Syscall stub>"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/primitives.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17protos/primitives.proto\x12\x0b\x61ngr.protos\"\x87\x05\n\rCodeReference\x12:\n\x0btarget_type\x18\x01 \x01(\x0e\x32%.angr.protos.CodeReference.TargetType\x12<\n\x0coperand_type\x18\x02 \x01(\x0e\x32&.angr.protos.CodeReference.OperandType\x12\x35\n\x08location\x18\x03 \x01(\x0e\x32#.angr.protos.CodeReference.Location\x12\n\n\x02\x65\x61\x18\x04 \x01(\x04\x12\x0c\n\x04mask\x18\x05 \x01(\x04\x12\x0c\n\x04name\x18\x06 \x01(\t\x12\x0f\n\x07\x64\x61ta_ea\x18\x07 \x01(\x04\x12\x10\n\x08\x62lock_ea\x18\x08 \x01(\x04\x12\x10\n\x08stmt_idx\x18\t \x01(\x05\x12\x13\n\x0boperand_idx\x18\n \x01(\x05\x12:\n\x08ref_type\x18\x0b \x01(\x0e\x32(.angr.protos.CodeReference.ReferenceType\"=\n\nTargetType\x12\x0e\n\nCodeTarget\x10\x00\x12\x0e\n\nDataTarget\x10\x01\x12\x0f\n\x0bStackTarget\x10\x02\"~\n\x0bOperandType\x12\x14\n\x10ImmediateOperand\x10\x00\x12\x11\n\rMemoryOperand\x10\x01\x12\x1d\n\x19MemoryDisplacementOperand\x10\x02\x12\x16\n\x12\x43ontrolFlowOperand\x10\x03\x12\x0f\n\x0bOffsetTable\x10\x04\"&\n\x08Location\x12\x0c\n\x08Internal\x10\x00\x12\x0c\n\x08\x45xternal\x10\x01\"0\n\rReferenceType\x12\n\n\x06offset\x10\x00\x12\x08\n\x04read\x10\x01\x12\t\n\x05write\x10\x02\"k\n\x0bInstruction\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\r\n\x05\x62ytes\x18\x02 \x01(\x0c\x12)\n\x05xrefs\x18\x03 \x01(\x0b\x32\x1a.angr.protos.CodeReference\x12\x16\n\x0elocal_noreturn\x18\x04 \x01(\x08\"`\n\x05\x42lock\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12.\n\x0cinstructions\x18\x02 \x01(\x0b\x32\x18.angr.protos.Instruction\x12\x0c\n\x04size\x18\x04 \x01(\r\x12\r\n\x05\x62ytes\x18\x05 \x01(\x0c\"\x95\x02\n\x10\x45xternalFunction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02\x65\x61\x18\x02 \x01(\x04\x12;\n\x02\x63\x63\x18\x03 \x01(\x0e\x32/.angr.protos.ExternalFunction.CallingConvention\x12\x12\n\nhas_return\x18\x04 \x01(\x08\x12\x11\n\tno_return\x18\x05 \x01(\x08\x12\x16\n\x0e\x61rgument_count\x18\x06 \x01(\x05\x12\x0f\n\x07is_weak\x18\x07 \x01(\x08\x12\x11\n\tprototype\x18\x08 \x01(\t\"G\n\x11\x43\x61llingConvention\x12\x11\n\rCallerCleanup\x10\x00\x12\x11\n\rCalleeCleanup\x10\x01\x12\x0c\n\x08\x46\x61stCall\x10\x02\"d\n\x10\x45xternalVariable\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\n\n\x02\x65\x61\x18\x02 \x01(\x04\x12\x0c\n\x04size\x18\x03 \x01(\r\x12\x0f\n\x07is_weak\x18\x04 \x01(\x08\x12\x17\n\x0fis_thread_local\x18\x05 \x01(\x08\"\xe5\x03\n\x04\x45\x64ge\x12\x0e\n\x06src_ea\x18\x01 \x01(\x04\x12\x0e\n\x06\x64st_ea\x18\x02 \x01(\x04\x12,\n\x08jumpkind\x18\x03 \x01(\x0e\x32\x1a.angr.protos.Edge.JumpKind\x12\x12\n\nis_outside\x18\x04 \x01(\x08\x12\x10\n\x08ins_addr\x18\x05 \x01(\x04\x12\x10\n\x08stmt_idx\x18\x06 \x01(\x03\x12)\n\x04\x64\x61ta\x18\x07 \x03(\x0b\x32\x1b.angr.protos.Edge.DataEntry\x1a+\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\"\xfe\x01\n\x08JumpKind\x12\x13\n\x0fUnknownJumpkind\x10\x00\x12\n\n\x06\x42oring\x10\x01\x12\x08\n\x04\x43\x61ll\x10\x02\x12\n\n\x06Return\x10\x03\x12\x0e\n\nFakeReturn\x10\x04\x12\x0b\n\x07Syscall\x10\x05\x12\x0f\n\x0bSys_syscall\x10\x06\x12\x0e\n\nSys_int128\x10\x07\x12\x0c\n\x08NoDecode\x10\x08\x12\n\n\x06\x45mWarn\x10\t\x12\x11\n\rSigFPE_IntDiv\x10\n\x12\x0b\n\x07SigTRAP\x10\x0b\x12\x0b\n\x07SigSEGV\x10\x0c\x12\x0b\n\x07MapFail\x10\r\x12\x0b\n\x07NoRedir\x10\x0e\x12\r\n\tClientReq\x10\x0f\x12\r\n\tException\x10\x10\".\n\nBlockGraph\x12 \n\x05\x65\x64ges\x18\x01 \x03(\x0b\x32\x11.angr.protos.Edgeb\x06proto3')



_CODEREFERENCE = DESCRIPTOR.message_types_by_name['CodeReference']
_INSTRUCTION = DESCRIPTOR.message_types_by_name['Instruction']
_BLOCK = DESCRIPTOR.message_types_by_name['Block']
_EXTERNALFUNCTION = DESCRIPTOR.message_types_by_name['ExternalFunction']
_EXTERNALVARIABLE = DESCRIPTOR.message_types_by_name['ExternalVariable']
_EDGE = DESCRIPTOR.message_types_by_name['Edge']
_EDGE_DATAENTRY = _EDGE.nested_types_by_name['DataEntry']
_BLOCKGRAPH = DESCRIPTOR.message_types_by_name['BlockGraph']
_CODEREFERENCE_TARGETTYPE = _CODEREFERENCE.enum_types_by_name['TargetType']
_CODEREFERENCE_OPERANDTYPE = _CODEREFERENCE.enum_types_by_name['OperandType']
_CODEREFERENCE_LOCATION = _CODEREFERENCE.enum_types_by_name['Location']
_CODEREFERENCE_REFERENCETYPE = _CODEREFERENCE.enum_types_by_name['ReferenceType']
_EXTERNALFUNCTION_CALLINGCONVENTION = _EXTERNALFUNCTION.enum_types_by_name['CallingConvention']
_EDGE_JUMPKIND = _EDGE.enum_types_by_name['JumpKind']
CodeReference = _reflection.GeneratedProtocolMessageType('CodeReference', (_message.Message,), {
  'DESCRIPTOR' : _CODEREFERENCE,
  '__module__' : 'protos.primitives_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.CodeReference)
  })
_sym_db.RegisterMessage(CodeReference)

Instruction = _reflection.GeneratedProtocolMessageType('Instruction', (_message.Message,), {
  'DESCRIPTOR' : _INSTRUCTION,
  '__module__' : 'protos.primitives_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.Instruction)
  })
_sym_db.RegisterMessage(Instruction)

Block = _reflection.GeneratedProtocolMessageType('Block', (_message.Message,), {
  'DESCRIPTOR' : _BLOCK,
  '__module__' : 'protos.primitives_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.Block)
  })
_sym_db.RegisterMessage(Block)

ExternalFunction = _reflection.GeneratedProtocolMessageType('ExternalFunction', (_message.Message,), {
  'DESCRIPTOR' : _EXTERNALFUNCTION,
  '__module__' : 'protos.primitives_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.ExternalFunction)
  })
_sym_db.RegisterMessage(ExternalFunction)

ExternalVariable = _reflection.GeneratedProtocolMessageType('ExternalVariable', (_message.Message,), {
  'DESCRIPTOR' : _EXTERNALVARIABLE,
  '__module__' : 'protos.primitives_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.ExternalVariable)
  })
_sym_db.RegisterMessage(ExternalVariable)

Edge = _reflection.GeneratedProtocolMessageType('Edge', (_message.Message,), {

  'DataEntry' : _reflection.GeneratedProtocolMessageType('DataEntry', (_message.Message,), {
    'DESCRIPTOR' : _EDGE_DATAENTRY,
    '__module__' : 'protos.primitives_pb2'
    # @@protoc_insertion_point(class_scope:angr.protos.Edge.DataEntry)
    })
  ,
  'DESCRIPTOR' : _EDGE,
  '__module__' : 'protos.primitives_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.Edge)
  })
_sym_db.RegisterMessage(Edge)
_sym_db.RegisterMessage(Edge.DataEntry)

BlockGraph = _reflection.GeneratedProtocolMessageType('BlockGraph', (_message.Message,), {
  'DESCRIPTOR' : _BLOCKGRAPH,
  '__module__' : 'protos.primitives_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.BlockGraph)
  })
_sym_db.RegisterMessage(BlockGraph)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EDGE_DATAENTRY._options = None
  _EDGE_DATAENTRY._serialized_options = b'8\001'
  _CODEREFERENCE._serialized_start=41
  _CODEREFERENCE._serialized_end=688
  _CODEREFERENCE_TARGETTYPE._serialized_start=409
  _CODEREFERENCE_TARGETTYPE._serialized_end=470
  _CODEREFERENCE_OPERANDTYPE._serialized_start=472
  _CODEREFERENCE_OPERANDTYPE._serialized_end=598
  _CODEREFERENCE_LOCATION._serialized_start=600
  _CODEREFERENCE_LOCATION._serialized_end=638
  _CODEREFERENCE_REFERENCETYPE._serialized_start=640
  _CODEREFERENCE_REFERENCETYPE._serialized_end=688
  _INSTRUCTION._serialized_start=690
  _INSTRUCTION._serialized_end=797
  _BLOCK._serialized_start=799
  _BLOCK._serialized_end=895
  _EXTERNALFUNCTION._serialized_start=898
  _EXTERNALFUNCTION._serialized_end=1175
  _EXTERNALFUNCTION_CALLINGCONVENTION._serialized_start=1104
  _EXTERNALFUNCTION_CALLINGCONVENTION._serialized_end=1175
  _EXTERNALVARIABLE._serialized_start=1177
  _EXTERNALVARIABLE._serialized_end=1277
  _EDGE._serialized_start=1280
  _EDGE._serialized_end=1765
  _EDGE_DATAENTRY._serialized_start=1465
  _EDGE_DATAENTRY._serialized_end=1508
  _EDGE_JUMPKIND._serialized_start=1511
  _EDGE_JUMPKIND._serialized_end=1765
  _BLOCKGRAPH._serialized_start=1767
  _BLOCKGRAPH._serialized_end=1813
# @@protoc_insertion_point(module_scope)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("ddraw.dll")
prototypes = \
    {
        #
        'DirectDrawEnumerateW': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpCallback", "lpContext"]),
        #
        'DirectDrawEnumerateA': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpCallback", "lpContext"]),
        #
        'DirectDrawEnumerateExW': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="HMONITOR")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3", "param4"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpCallback", "lpContext", "dwFlags"]),
        #
        'DirectDrawEnumerateExA': SimTypeFunction([SimTypePointer(SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeBottom(label="HMONITOR")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0", "param1", "param2", "param3", "param4"]), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpCallback", "lpContext", "dwFlags"]),
        #
        'DirectDrawCreate': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeBottom(label="IDirectDraw"), offset=0), SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpGUID", "lplpDD", "pUnkOuter"]),
        #
        'DirectDrawCreateEx': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpGuid", "lplpDD", "iid", "pUnkOuter"]),
        #
        'DirectDrawCreateClipper': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="IDirectDrawClipper"), offset=0), SimTypeBottom(label="IUnknown")], SimTypeInt(signed=True, label="Int32"), arg_names=["dwFlags", "lplpDDClipper", "pUnkOuter"]),
    }

lib.set_prototypes(prototypes)



import claripy


class PropagatorLoadCallback:
    """
    Implement the load callback for Propagator that should be used across all jump table resolvers.
    """

    def __init__(self, project):
        self.project = project

    def propagator_load_callback(self, addr, size) -> bool:  # pylint:disable=unused-argument
        # only allow loading if the address falls into a read-only region
        if isinstance(addr, claripy.ast.BV) and addr.op == "BVV":
            addr_v = addr.args[0]
            section = self.project.loader.find_section_containing(addr_v)
            if section is not None:
                return section.is_readable and not section.is_writable
            segment = self.project.loader.find_segment_containing(addr_v)
            if segment is not None:
                return segment.is_readable and not segment.is_writable
        return False



import angr


class fgetc(angr.SimProcedure):
    # pylint:disable=arguments-differ
    def run(self, stream, simfd=None):
        if simfd is None:
            fileno = angr.SIM_PROCEDURES["posix"]["fileno"]
            fd = self.inline_call(fileno, stream).ret_expr
            simfd = self.state.posix.get_fd(fd)

        if simfd is None:
            return -1

        (
            data,
            real_length,
        ) = simfd.read_data(1)
        return self.state.solver.If(real_length == 0, -1, data.zero_extend(self.arch.sizeof["int"] - 8))


getc = fgetc
fgetc_unlocked = fgetc
getc_unlocked = fgetc



import angr

import logging

l = logging.getLogger(name=__name__)


class atol(angr.SimProcedure):
    # pylint:disable=arguments-differ
    def run(self, s):
        strtol = angr.SIM_PROCEDURES["libc"]["strtol"]
        return strtol.strtol_inner(s, self.state, self.state.memory, 10, True)[1]



import logging
import weakref
import itertools
from contextlib import contextmanager
import gc
import networkx

import claripy

from .misc.picklable_lock import PicklableRLock

l = logging.getLogger(name=__name__)


class StateHierarchy:
    """
    The state hierarchy holds weak references to SimStateHistory objects in a directed acyclic graph. It is useful
    for queries about a state's ancestry, notably "what is the best ancestor state for a merge among these states" and
    "what is the most recent unsatisfiable state while using LAZY_SOLVES"
    """

    def __init__(self):
        # The New Order
        self._graph = networkx.DiGraph()
        self._leaves = set()  # nodes with no children
        self._twigs = set()  # nodes with one child
        self._weakref_cache = {}  # map from object id to weakref
        self._reverse_weakref_cache = {}  # map from weakref to object id
        self._pending_cleanup = set()
        self._defer_cleanup = False
        self._lock = PicklableRLock()

    def __getstate__(self):
        gc.collect()
        histories = [h() for h in networkx.algorithms.dfs_postorder_nodes(self._graph)]
        return (histories,)

    def __setstate__(self, s):
        self._graph = networkx.DiGraph()
        self._leaves = set()
        self._twigs = set()
        self._weakref_cache = {}
        self._reverse_weakref_cache = {}

        nodes = s[0]
        for n in nodes:
            self.add_history(n)

    def get_ref(self, obj):
        if id(obj) not in self._weakref_cache:
            ref = weakref.ref(obj, self.dead_ref)
            self._weakref_cache[id(obj)] = ref
            self._reverse_weakref_cache[ref] = id(obj)
            return ref
        else:
            return self._weakref_cache[id(obj)]

    def dead_ref(self, ref):
        if self._defer_cleanup:
            self._pending_cleanup.add(ref)
        else:
            self._cleanup_ref(ref)

    def _cleanup_ref(self, ref):
        if ref not in self._reverse_weakref_cache:
            l.error("Cleaning mystery weakref %s", ref)
            return

        self._remove_history(ref)

        del self._weakref_cache[self._reverse_weakref_cache[ref]]
        del self._reverse_weakref_cache[ref]

    @contextmanager
    def defer_cleanup(self):
        old_defer, self._defer_cleanup = self._defer_cleanup, True
        try:
            yield
        finally:
            self._defer_cleanup = old_defer
            if not self._defer_cleanup:
                toclean = list(self._pending_cleanup)
                self._pending_cleanup.clear()
                for ref in toclean:
                    self._cleanup_ref(ref)

    #
    # Graph management
    #

    def _remove_history(self, h):
        with self.defer_cleanup(), self._lock:
            try:
                predecessors = self._graph.predecessors(h)
                successors = self._graph.successors(h)

                for p, s in itertools.product(predecessors, successors):
                    self._graph.add_edge(p, s)

                self._graph.remove_node(h)
            except networkx.NetworkXError:
                pass

            self._leaves.discard(h)
            self._twigs.discard(h)
            hh = h()
            if hh is not None:
                hh.demote()

    def add_state(self, s):
        h = s.history
        self.add_history(h)

    def add_history(self, h):
        with self.defer_cleanup(), self._lock:
            cur_node = self.get_ref(h)
            self._graph.add_node(cur_node)
            if h.parent is not None:
                prev_node = self.get_ref(h.parent)
                self._graph.add_edge(prev_node, cur_node)

                self._leaves.discard(prev_node)
                if len(list(self._graph.successors(prev_node))) == 1:
                    self._twigs.add(prev_node)
                else:
                    self._twigs.discard(prev_node)

            self._leaves.add(cur_node)

    def simplify(self):
        tw, self._twigs = self._twigs, set()
        for h in tw:
            self._remove_history(h)

    def full_simplify(self):
        with self.defer_cleanup(), self._lock:
            for h in self._graph.nodes():
                if self._graph.out_degree(h) == 1:
                    self._remove_history(h)

    def lineage(self, h):
        """
        Returns the lineage of histories leading up to `h`.
        """

        with self.defer_cleanup(), self._lock:
            lineage = []

            predecessors = list(self._graph.predecessors(h))
            while len(predecessors):
                lineage.append(predecessors[0])
                predecessors = list(self._graph.predecessors(predecessors[0]))

            lineage.reverse()
            return lineage

    def all_successors(self, h):
        with self.defer_cleanup(), self._lock:
            nodes = list(networkx.algorithms.dfs_postorder_nodes(self._graph, h))[:-1]
            nodes.reverse()
            return nodes

    def history_successors(self, h):
        with self.defer_cleanup(), self._lock:
            return [ref() for ref in self._graph.successors(self.get_ref(h))]

    def history_predecessors(self, h):
        with self.defer_cleanup(), self._lock:
            return [ref() for ref in self._graph.predecessors(self.get_ref(h))]

    def history_contains(self, h):
        with self.defer_cleanup(), self._lock:
            return self.get_ref(h) in self._graph

    #
    # LAZY_SOLVES support
    #

    def _find_root_unreachable(self, h):
        lineage = self.lineage(h)
        if len(lineage) == 0 or lineage[-1]().reachable():
            return h

        good = 0
        bad = len(lineage) - 1

        while True:
            l.debug("... looking between %d and %d in %d states", good, bad, len(lineage))
            cur = (bad + good) // 2

            if cur in (good, bad):
                if lineage[bad]().reachable():
                    bad += 1

                root = lineage[bad]
                l.debug("... returning %d (%s)", bad, root)
                return root
            elif lineage[cur]().reachable():
                l.debug("... %d is reachable", cur)
                good = cur
            else:
                l.debug("... %d is unreachable", bad)
                bad = cur

    def _prune_subtree(self, h):
        with self.defer_cleanup(), self._lock:
            ph = list(self._graph.predecessors(h))
            if len(ph) == 1 and len(list(self._graph.successors(ph[0]))) == 1:
                self._twigs.add(ph[0])

            all_children = list(networkx.algorithms.dfs_postorder_nodes(self._graph, h))
            for n in all_children:
                n()._satisfiable = False
                try:
                    if n().state is not None:
                        n().state.add_constraints(claripy.false)
                except ReferenceError:
                    pass
            self._graph.remove_nodes_from(all_children)

    def unreachable_state(self, state):
        self.unreachable_history(state.history)

    def unreachable_history(self, h):
        href = self.get_ref(h)

        with self._lock:
            try:
                l.debug("Pruning tree given unreachable %s", h)
                root = self._find_root_unreachable(href)
            except networkx.NetworkXError:
                l.debug("... not present in graph")
            else:
                l.debug("... root is %s", root)
                self._prune_subtree(root)

    #
    # Smart merging support
    #

    def most_mergeable(self, states):
        """
        Find the "most mergeable" set of states from those provided.

        :param states: a list of states
        :returns: a tuple of: (list of states to merge, those states' common history, list of states to not merge yet)
        """

        with self.defer_cleanup(), self._lock:
            histories = {self.get_ref(s.history) for s in states}

            for n in networkx.algorithms.dfs_postorder_nodes(self._graph):
                intersection = histories.intersection(self.all_successors(n))
                if len(intersection) > 1:
                    return (
                        [s for s in states if self.get_ref(s.history) in intersection],
                        n(),
                        [s for s in states if self.get_ref(s.history) not in intersection],
                    )

            # didn't find any?
            return set(), None, states



# pylint:disable=unsubscriptable-object
from typing import Set, Optional, Tuple, Any, Union, TYPE_CHECKING

from ...utils.cowdict import DefaultChainMapCOW
from ...code_location import CodeLocation

if TYPE_CHECKING:
    from .definition import Definition


class Uses:
    """
    Describes uses (including the use location and the use expression) for definitions.
    """

    __slots__ = ("_uses_by_definition", "_uses_by_location")

    def __init__(
        self,
        uses_by_definition: Optional[DefaultChainMapCOW] = None,
        uses_by_location: Optional[DefaultChainMapCOW] = None,
    ):
        self._uses_by_definition: DefaultChainMapCOW["Definition", Set[Tuple[CodeLocation, Optional[Any]]]] = (
            DefaultChainMapCOW(set, collapse_threshold=25) if uses_by_definition is None else uses_by_definition
        )
        self._uses_by_location: DefaultChainMapCOW[CodeLocation, Set[Tuple["Definition", Optional[Any]]]] = (
            DefaultChainMapCOW(set, collapse_threshold=25) if uses_by_location is None else uses_by_location
        )

    def add_use(self, definition: "Definition", codeloc: CodeLocation, expr: Optional[Any] = None):
        """
        Add a use for a given definition.

        :param definition:  The definition that is used.
        :param codeloc:     The code location where the use occurs.
        :param expr:        The expression that uses the specified definition at this location.
        """
        self._uses_by_definition[definition].add((codeloc, expr))
        self._uses_by_location[codeloc].add((definition, expr))

    def get_uses(self, definition: "Definition") -> Set[CodeLocation]:
        """
        Retrieve the uses of a given definition.

        :param definition: The definition for which we get the uses.
        """
        return {codeloc for codeloc, _ in self._uses_by_definition.get(definition, set())}

    def get_uses_with_expr(self, definition: "Definition") -> Set[Tuple[CodeLocation, Optional[Any]]]:
        """
        Retrieve the uses and the corresponding expressions of a given definition.

        :param definition: The definition for which we get the uses and the corresponding expressions.
        """
        return self._uses_by_definition.get(definition, set())

    def remove_use(self, definition: "Definition", codeloc: "CodeLocation", expr: Optional[Any] = None) -> None:
        """
        Remove one use of a given definition.

        :param definition:  The definition of which to remove the uses.
        :param codeloc:     The code location where the use is.
        :param expr:        The expression that uses the definition at the given location.
        :return:            None
        """
        if definition in self._uses_by_definition:
            if codeloc in self._uses_by_definition[definition]:
                if expr is None:
                    for codeloc_, expr_ in list(self._uses_by_definition[definition]):
                        if codeloc_ == codeloc:
                            self._uses_by_definition[definition].remove((codeloc_, expr_))
                else:
                    self._uses_by_definition[definition].remove((codeloc, expr))

        if codeloc in self._uses_by_location:
            for item in list(self._uses_by_location[codeloc]):
                if item[0] == definition:
                    self._uses_by_location[codeloc].remove(item)

    def remove_uses(self, definition: "Definition"):
        """
        Remove all uses of a given definition.

        :param definition:  The definition of which to remove the uses.
        :return:            None
        """
        if definition in self._uses_by_definition:
            codeloc_and_ids = self._uses_by_definition[definition]
            del self._uses_by_definition[definition]

            for codeloc, _ in codeloc_and_ids:
                for item in list(self._uses_by_location[codeloc]):
                    if item[0] == definition:
                        self._uses_by_location[codeloc].remove(item)

    def get_uses_by_location(
        self, codeloc: CodeLocation, exprs: bool = False
    ) -> Union[Set["Definition"], Set[Tuple["Definition", Optional[Any]]]]:
        """
        Retrieve all definitions that are used at a given location.

        :param codeloc: The code location.
        :return:        A set of definitions that are used at the given location.
        """
        if exprs:
            return self._uses_by_location.get(codeloc, set())
        return {item[0] for item in self._uses_by_location.get(codeloc, set())}

    def get_uses_by_insaddr(
        self, ins_addr: int, exprs: bool = False
    ) -> Union[Set["Definition"], Set[Tuple["Definition", Optional[Any]]]]:
        """
        Retrieve all definitions that are used at a given location specified by the instruction address.

        :param ins_addr:    The instruction address.
        :return:            A set of definitions that are used at the given location.
        """

        all_uses = set()
        for codeloc, uses in self._uses_by_location.items():
            if codeloc.ins_addr == ins_addr:
                all_uses |= uses

        if exprs:
            return all_uses
        return {item[0] for item in all_uses}

    def copy(self) -> "Uses":
        """
        Copy the instance.

        :return:    Return a new <Uses> instance containing the same data.
        """
        u = Uses(
            uses_by_definition=self._uses_by_definition.copy(),
            uses_by_location=self._uses_by_location.copy(),
        )

        return u

    def merge(self, other: "Uses") -> bool:
        """
        Merge an instance of <Uses> into the current instance.

        :param other: The other <Uses> from which the data will be added to the current instance.
        :return: True if any merge occurred, False otherwise
        """
        merge_occurred = False

        for k, v in other._uses_by_definition.items():
            if k not in self._uses_by_definition:
                self._uses_by_definition[k] = v
                merge_occurred = True
            elif not v.issubset(self._uses_by_definition[k]):
                merge_occurred = True
                self._uses_by_definition[k] |= v

        for k, v in other._uses_by_location.items():
            if k not in self._uses_by_location:
                self._uses_by_location[k] = v
                merge_occurred = True
            elif not v.issubset(self._uses_by_location[k]):
                merge_occurred = True
                self._uses_by_location[k] |= v

        return merge_occurred



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("wnvapi.dll")
prototypes = \
    {
        #
        'WnvOpen': SimTypeFunction([], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)),
        #
        'WnvRequestNotification': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"Header": SimStruct({"MajorVersion": SimTypeChar(label="Byte"), "MinorVersion": SimTypeChar(label="Byte"), "Size": SimTypeInt(signed=False, label="UInt32")}, name="WNV_OBJECT_HEADER", pack=False, align=None), "NotificationType": SimTypeInt(signed=False, label="WNV_NOTIFICATION_TYPE"), "PendingNotifications": SimTypeInt(signed=False, label="UInt32"), "Buffer": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="WNV_NOTIFICATION_PARAM", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Internal": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "InternalHigh": SimTypePointer(SimTypeInt(signed=False, label="UInt"), label="UIntPtr", offset=0), "Anonymous": SimUnion({"Anonymous": SimStruct({"Offset": SimTypeInt(signed=False, label="UInt32"), "OffsetHigh": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous_e__Struct", pack=False, align=None), "Pointer": SimTypePointer(SimTypeBottom(label="Void"), offset=0)}, name="<anon>", label="None"), "hEvent": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="OVERLAPPED", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["WnvHandle", "NotificationParam", "Overlapped", "BytesTransferred"]),
    }

lib.set_prototypes(prototypes)
